"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ApiKeyDestroy = exports.ApiKeyView = exports.ApiKeyEdit = exports.ApiKeyCreate = exports.ApiKeysList = void 0;
const authenticatedAction_1 = require("../classes/actions/authenticatedAction");
const ApiKey_1 = require("../models/ApiKey");
const apiData_1 = require("../modules/apiData");
class ApiKeysList extends authenticatedAction_1.AuthenticatedAction {
    constructor() {
        super(...arguments);
        this.name = "apiKeys:list";
        this.description = "list all the apiKeys";
        this.outputExample = {};
        this.permission = { topic: "apiKey", mode: "read" };
        this.inputs = {
            limit: { required: true, default: 100, formatter: apiData_1.APIData.ensureNumber },
            offset: { required: true, default: 0, formatter: apiData_1.APIData.ensureNumber },
        };
    }
    async runWithinTransaction({ params }) {
        const total = await ApiKey_1.ApiKey.count();
        const apiKeys = await ApiKey_1.ApiKey.findAll({
            limit: params.limit,
            offset: params.offset,
        });
        return {
            total,
            apiKeys: await Promise.all(apiKeys.map((apiKey) => apiKey.apiData())),
        };
    }
}
exports.ApiKeysList = ApiKeysList;
class ApiKeyCreate extends authenticatedAction_1.AuthenticatedAction {
    constructor() {
        super(...arguments);
        this.name = "apiKey:create";
        this.description = "create an apiKey";
        this.outputExample = {};
        this.permission = { topic: "apiKey", mode: "write" };
        this.inputs = {
            name: { required: true },
            permissionAllRead: { required: false },
            permissionAllWrite: { required: false },
            permissions: { required: false, formatter: apiData_1.APIData.ensureArray },
        };
    }
    async runWithinTransaction({ params }) {
        const apiKey = new ApiKey_1.ApiKey(params);
        await apiKey.save();
        if (params.permissions)
            await apiKey.setPermissions(params.permissions);
        return { apiKey: await apiKey.apiData() };
    }
}
exports.ApiKeyCreate = ApiKeyCreate;
class ApiKeyEdit extends authenticatedAction_1.AuthenticatedAction {
    constructor() {
        super(...arguments);
        this.name = "apiKey:edit";
        this.description = "edit an apiKey";
        this.outputExample = {};
        this.permission = { topic: "apiKey", mode: "write" };
        this.inputs = {
            id: { required: true },
            name: { required: false },
            permissionAllRead: { required: false },
            permissionAllWrite: { required: false },
            disabledPermissionAllRead: { required: false },
            disabledPermissionAllWrite: { required: false },
            permissions: { required: false, formatter: apiData_1.APIData.ensureArray },
        };
    }
    async runWithinTransaction({ params }) {
        const apiKey = await ApiKey_1.ApiKey.findById(params.id);
        const updateParams = Object.assign({}, params);
        if (params.disabledPermissionAllRead) {
            updateParams.permissionAllRead = null;
        }
        if (params.disabledPermissionAllWrite) {
            updateParams.permissionAllWrite = null;
        }
        await apiKey.update(updateParams);
        if (params.permissions)
            await apiKey.setPermissions(params.permissions);
        return { apiKey: await apiKey.apiData() };
    }
}
exports.ApiKeyEdit = ApiKeyEdit;
class ApiKeyView extends authenticatedAction_1.AuthenticatedAction {
    constructor() {
        super(...arguments);
        this.name = "apiKey:view";
        this.description = "view an apiKey";
        this.outputExample = {};
        this.permission = { topic: "apiKey", mode: "read" };
        this.inputs = {
            id: { required: true },
        };
    }
    async runWithinTransaction({ params }) {
        const apiKey = await ApiKey_1.ApiKey.findById(params.id);
        return { apiKey: await apiKey.apiData() };
    }
}
exports.ApiKeyView = ApiKeyView;
class ApiKeyDestroy extends authenticatedAction_1.AuthenticatedAction {
    constructor() {
        super(...arguments);
        this.name = "apiKey:destroy";
        this.description = "destroy an apiKey";
        this.outputExample = {};
        this.permission = { topic: "apiKey", mode: "write" };
        this.inputs = {
            id: { required: true },
        };
    }
    async runWithinTransaction({ params, }) {
        const apiKey = await ApiKey_1.ApiKey.findById(params.id);
        await apiKey.destroy();
        return { success: true };
    }
}
exports.ApiKeyDestroy = ApiKeyDestroy;
