"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.StatusCLI = void 0;
const cli_1 = require("../modules/cli");
const plugin_1 = require("../modules/plugin");
const group_1 = require("../modules/ops/group");
const actionhero_1 = require("actionhero");
const cls_1 = require("../modules/cls");
const status_1 = require("../modules/status");
class StatusCLI extends actionhero_1.CLI {
    constructor() {
        super();
        this.name = "status";
        this.description = "Display the status of your Grouparoo cluster";
        this.preInitialize = () => {
            cli_1.GrouparooCLI.setGrouparooRunMode(this);
            cli_1.GrouparooCLI.setNextDevelopmentMode();
        };
        cli_1.GrouparooCLI.timestampOption(this);
        cli_1.GrouparooCLI.jsonOption(this);
    }
    async run({ params }) {
        if (!params.json)
            cli_1.GrouparooCLI.logCLI(this.name);
        await cls_1.CLS.wrap(async () => {
            const { value: clusterName } = await plugin_1.plugin.readSetting("core", "cluster-name");
            await status_1.Status.setAll();
            const samples = await status_1.Status.get();
            const { groups, newestMembersAdded } = await group_1.GroupOps.newestGroupMembers(100);
            const groupsStatus = {};
            for (const i in groups) {
                const group = groups[i];
                const additionTime = newestMembersAdded[group.id]
                    ? new Date(newestMembersAdded[group.id]).toISOString()
                    : "Never";
                const groupMembersCount = await group.$count("groupMembers");
                groupsStatus[group.name] = [
                    `${groupMembersCount} members`,
                    "newest member: " + additionTime,
                ];
            }
            const overview = {
                ClusterName: [clusterName, actionhero_1.env ? `${actionhero_1.env}` : undefined],
                TotalRecords: [
                    getLatestMetric(samples, "GrouparooRecord", "totals").count,
                ],
                TotalGroups: [getLatestMetric(samples, "Group", "totals").count],
            };
            const pendingItems = [];
            const pendingRuns = [];
            const pendingDeletions = [];
            for (const topic in samples) {
                for (const collection in samples[topic]) {
                    const metrics = samples[topic][collection];
                    const { metric } = metrics[metrics.length - 1];
                    if (metric.collection === "pending") {
                        pendingItems.push({
                            [metric.topic]: [metric.count],
                        });
                    }
                    if (metric.collection === "deleted") {
                        pendingDeletions.push({ [metric.topic]: [metric.count] });
                    }
                    if (metric.topic === "Run" &&
                        metric.collection === "percentComplete") {
                        pendingRuns.push({
                            [metric.value]: [
                                `${metric.count}%${metric.metadata ? ` (${metric.metadata})` : ""}`,
                            ],
                        });
                    }
                }
            }
            const data = [
                { header: "Overview", status: overview },
                { header: "Groups", status: groupsStatus },
                {
                    header: "Pending Items",
                    status: pendingItems.reduce((s, arr) => Object.assign(s, arr), {}),
                },
                {
                    header: "Active Runs",
                    status: pendingRuns.reduce((s, arr) => Object.assign(s, arr), {}),
                },
                {
                    header: "Pending Deletions",
                    status: pendingDeletions.reduce((s, arr) => Object.assign(s, arr), {}),
                },
            ];
            if (params.json) {
                const jsonData = {};
                data.map((collection) => {
                    jsonData[collection.header] = collection.status;
                });
                cli_1.GrouparooCLI.logger.log(JSON.stringify(jsonData));
            }
            else {
                cli_1.GrouparooCLI.logger.status("Cluster Status", data);
            }
        });
        return true;
    }
}
exports.StatusCLI = StatusCLI;
function getLatestMetric(samples, topic, collection) {
    return samples[topic][collection][samples[topic][collection].length - 1]
        .metric;
}
