"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Plugins = void 0;
const actionhero_1 = require("actionhero");
const plugin_1 = require("../modules/plugin");
const App_1 = require("../models/App");
const pluginDetails_1 = require("../modules/pluginDetails");
const cls_1 = require("../modules/cls");
const runMode_1 = require("../modules/runMode");
class Plugins extends actionhero_1.Initializer {
    constructor() {
        super();
        this.name = "plugins";
        this.loadPriority = 999;
        this.startPriority = 1;
        this.stopPriority = 900;
    }
    async initialize() {
        actionhero_1.api.plugins = {
            plugins: [],
            persistentConnections: {},
            validate: this.validatePlugin,
            register: this.registerPlugin,
            announcePlugins: this.announcePlugins,
        };
        this.checkPluginEnvironmentVariables();
        // --- Add the core plugins --- //
        plugin_1.plugin.registerPlugin({
            name: "@grouparoo/core",
            icon: "/public/@grouparoo/logo.png",
        });
        // --- Add the UI pLugin --- //
        const pluginManifest = (0, pluginDetails_1.getPluginManifest)();
        const uiPlugins = pluginManifest.plugins
            .filter((p) => p.name.match(/^@grouparoo\/ui-/))
            .sort((p) => (p.name.match("config") ? 1 : -1));
        if (uiPlugins) {
            const relevantUiPlugin = uiPlugins[(0, runMode_1.getGrouparooRunMode)() === "cli:config" ? uiPlugins.length - 1 : 0];
            if (relevantUiPlugin) {
                this.registerPlugin({
                    name: relevantUiPlugin.name,
                    icon: "/public/@grouparoo/logo.png",
                });
                process.env.GROUPAROO_UI_EDITION = relevantUiPlugin.name.replace("@grouparoo/ui-", "");
            }
        }
    }
    async start() {
        actionhero_1.api.plugins.announcePlugins();
    }
    async stop() {
        await cls_1.CLS.wrap(async () => {
            for (const id in actionhero_1.api.plugins.persistentConnections) {
                const app = await App_1.App.findById(id).catch(() => null);
                await (app === null || app === void 0 ? void 0 : app.disconnect());
            }
            await actionhero_1.utils.sleep(100);
        });
    }
    validatePlugin(plugin) {
        let errors = [];
        if (!plugin.name) {
            errors.push(`name is required for a Grouparoo Plugin`);
        }
        if (actionhero_1.api.plugins.plugins.map((p) => p.name).includes(plugin.name)) {
            errors.push(`a plugin named ${plugin.name} is already registered`);
        }
        if (plugin.apps) {
            plugin === null || plugin === void 0 ? void 0 : plugin.apps.forEach((app) => {
                if (!app.name) {
                    errors.push(`app.name is required for a Grouparoo Plugin App`);
                }
                if (actionhero_1.api.plugins.plugins
                    .map((p) => { var _a; return (_a = p.apps) === null || _a === void 0 ? void 0 : _a.map((app) => app.name); })
                    .flat()
                    .includes(plugin.name)) {
                    errors.push(`a plugin app named ${app.name} is already registered`);
                }
                if (!app.options) {
                    errors.push(`app.options is required for a Grouparoo Plugin App`);
                }
                if (!app.methods.test) {
                    errors.push(`app.methods.test is required for a Grouparoo Plugin App`);
                }
                if ((app.methods.connect && !app.methods.disconnect) ||
                    (!app.methods.connect && app.methods.disconnect)) {
                    errors.push(`both app.methods.connect and app.methods.disconnect are required for a Grouparoo Plugin App`);
                }
            });
        }
        if (plugin.connections) {
            plugin.connections.forEach((connection) => {
                if (!connection.name) {
                    errors.push(`connection.name is required for a Grouparoo Plugin Connection`);
                }
                if (actionhero_1.api.plugins.plugins
                    .map((p) => { var _a; return (_a = p.connections) === null || _a === void 0 ? void 0 : _a.map((conn) => conn.name); })
                    .flat()
                    .includes(connection.name)) {
                    errors.push(`a plugin connection named ${connection.name} is already registered`);
                }
                if (!connection.apps || connection.apps.length < 1) {
                    errors.push(`connection.apps is required for a Grouparoo Plugin Connection (${connection.name})`);
                }
                if (!connection.description) {
                    errors.push(`connection.description is required for a Grouparoo Plugin Connection (${connection.name})`);
                }
                if (!connection.direction) {
                    errors.push(`connection.direction is required for a Grouparoo Plugin Connection (${connection.name})`);
                }
                if (!connection.options) {
                    errors.push(`connection.options is required for a Grouparoo Plugin Connection (${connection.name})`);
                }
                if (!["import", "export"].includes(connection.direction)) {
                    errors.push(`connection.direction must be either import or export for a Grouparoo Plugin Connection (${connection.name})`);
                }
                if (connection.direction === "import" &&
                    !connection.methods.importRecords &&
                    !connection.methods.recordProperty &&
                    !connection.methods.recordProperties) {
                    errors.push(`import connections must provide a way to import records (${connection.name})`);
                }
                if (connection.direction === "import" &&
                    connection.methods.importRecords &&
                    connection.supportIncrementalSchedule === undefined) {
                    errors.push(`import connections that support schedules must declare supportIncrementalSchedule`);
                }
                if (connection.direction === "export" &&
                    !connection.methods.exportRecord &&
                    !connection.methods.exportRecords) {
                    errors.push(`export connections must provide either connection.methods.exportRecord or connection.methods.exportRecords (${connection.name})`);
                }
            });
        }
        if (errors.length > 0) {
            throw new Error(`Plugin validation failure for ${plugin.name ? `"${plugin.name}"` : "plugin"}: ${errors.join(", ")}`);
        }
        return true;
    }
    registerPlugin(plugin, validate = true) {
        if (validate)
            actionhero_1.api.plugins.validate(plugin);
        actionhero_1.api.plugins.plugins.push(plugin);
    }
    announcePlugins() {
        (0, actionhero_1.log)(`active plugins: ${actionhero_1.api.plugins.plugins.map((p) => p.name).join(", ")}`);
    }
    /*
     * ensure that the ENV variables for the plugins are set
     */
    checkPluginEnvironmentVariables() {
        const pluginManifest = (0, pluginDetails_1.getPluginManifest)();
        pluginManifest.plugins.forEach((plugin) => {
            var _a, _b, _c, _d;
            if ((_b = (_a = plugin === null || plugin === void 0 ? void 0 : plugin.grouparoo) === null || _a === void 0 ? void 0 : _a.env) === null || _b === void 0 ? void 0 : _b.api) {
                plugin.grouparoo.env.api.forEach((e) => {
                    if (!process.env[e]) {
                        (0, actionhero_1.log)(`Plugin ${plugin.name} requires environment variable ${e} to be set`, "warning");
                    }
                });
            }
            if ((_d = (_c = plugin === null || plugin === void 0 ? void 0 : plugin.grouparoo) === null || _c === void 0 ? void 0 : _c.env) === null || _d === void 0 ? void 0 : _d.web) {
                plugin.grouparoo.env.web.forEach((e) => {
                    if (!process.env[e]) {
                        (0, actionhero_1.log)(`plugin ${plugin.name} requires environment variable ${e} to be set`, "warning");
                    }
                });
            }
        });
    }
}
exports.Plugins = Plugins;
