"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.CloudClient = exports.packageConfig = exports.CloudError = void 0;
const os_1 = __importDefault(require("os"));
const path_1 = __importDefault(require("path"));
const tar_1 = __importDefault(require("tar"));
const fs_1 = __importDefault(require("fs"));
const form_data_1 = __importDefault(require("form-data"));
const isomorphic_fetch_1 = __importDefault(require("isomorphic-fetch"));
const fs_extra_1 = require("fs-extra");
const actionhero_1 = require("actionhero");
class CloudError extends Error {
    constructor({ code, message }) {
        super(message);
        this.code = code;
    }
}
exports.CloudError = CloudError;
async function packageConfig(projectPath, configDirPath, tarballPath) {
    const tempDir = await (0, fs_extra_1.mkdtemp)(path_1.default.join(os_1.default.tmpdir(), "grouparoo-pack-"));
    await (0, fs_extra_1.copy)(path_1.default.join(projectPath, "package.json"), path_1.default.join(tempDir, "package.json"));
    await (0, fs_extra_1.copy)(configDirPath, path_1.default.join(tempDir, "config"), {
        filter: (src) => {
            const ext = path_1.default.extname(src);
            return !ext || [".json", ".js"].includes(ext);
        },
    });
    await tar_1.default.create({
        file: tarballPath,
        cwd: tempDir,
        gzip: true,
        strict: true,
    }, ["."]);
    await (0, fs_extra_1.remove)(tempDir);
    return tarballPath;
}
exports.packageConfig = packageConfig;
const maxAttempts = 5;
class CloudClient {
    constructor(projectId, token) {
        var _a;
        this.projectId = projectId;
        this.token = token;
        this.baseUrl =
            (_a = process.env.GROUPAROO_CLOUD_API_URL) !== null && _a !== void 0 ? _a : "https://cloud.grouparoo.com";
    }
    async request(url, options, attempts = 0) {
        const fetchUrl = new URL(url, this.baseUrl);
        fetchUrl.searchParams.append("apiToken", this.token);
        // when retrying a form based on a readStream, we need to restart the stream on every attempt
        if (typeof (options === null || options === void 0 ? void 0 : options._buildFormData) === "function") {
            const formData = options._buildFormData.bind(this)();
            //@ts-ignore typings are incorrectly not allowing FormData to be set as body
            options.body = formData;
            options.headers = formData.getHeaders();
        }
        try {
            const optionsToSend = {
                ...options,
                _buildFormData: undefined,
            };
            const res = await (0, isomorphic_fetch_1.default)(fetchUrl.toString(), optionsToSend);
            const data = (await res.json());
            if (res.status !== 200) {
                if (data.error)
                    throw new CloudError(data.error);
                throw new Error(await res.text());
            }
            return data;
        }
        catch (error) {
            if (((error === null || error === void 0 ? void 0 : error.code) === "ETIMEDOUT" || error.toString().match("ETIMEDOUT")) &&
                attempts < maxAttempts - 1) {
                await actionhero_1.utils.sleep(100);
                return this.request(url, options, attempts + 1);
            }
            else
                throw error;
        }
    }
    async createConfiguration(tarballPath, toApply, message, externalUrl) {
        function _buildFormData() {
            const formData = new form_data_1.default();
            formData.append("_file", fs_1.default.createReadStream(tarballPath));
            formData.append("projectId", this.projectId);
            formData.append("toApply", toApply.toString());
            if (message)
                formData.append("message", message);
            if (externalUrl)
                formData.append("externalUrl", externalUrl);
            return formData;
        }
        const { configuration } = await this.request("/api/v1/configuration", {
            method: "POST",
            _buildFormData,
        });
        return configuration;
    }
    async getConfiguration(configurationId) {
        const { configuration } = await this.request(`/api/v1/configuration/${configurationId}`);
        return configuration;
    }
    async getJob(jobId) {
        const { job } = await this.request(`/api/v1/job/${jobId}`);
        return job;
    }
}
exports.CloudClient = CloudClient;
