"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.FilterHelper = void 0;
const Filter_1 = require("../models/Filter");
const Property_1 = require("../models/Property");
const Schedule_1 = require("../models/Schedule");
const Source_1 = require("../models/Source");
const Mapping_1 = require("../models/Mapping");
const Option_1 = require("../models/Option");
const modelName_1 = require("./modelName");
var FilterHelper;
(function (FilterHelper) {
    FilterHelper.deprecatedFilters = {
        equals: "eq",
        "does not equal": "ne",
        "greater than": "gt",
        "greater than or equal to": "gte",
        "less than": "lt",
        "less than or equal to": "lte",
        contains: "substring",
        "does not contain": "notSubstring",
    };
    /**
     * Get the existing filters for an instance
     *
     */
    async function getFilters(instance) {
        const filtersWithCol = [];
        const filters = instance.filters
            ? instance.filters.sort((a, b) => a.position - b.position)
            : await Filter_1.Filter.findAll({
                where: {
                    ownerId: instance.id,
                    ownerType: (0, modelName_1.modelName)(instance),
                },
                order: [["position", "asc"]],
            });
        if (!instance.filters)
            instance.filters = filters;
        for (const i in filters) {
            const filter = filters[i];
            filtersWithCol.push({
                key: filter.key,
                op: filter.op,
                match: filter.match,
                relativeMatchNumber: filter.relativeMatchNumber,
                relativeMatchUnit: filter.relativeMatchUnit,
                relativeMatchDirection: filter.relativeMatchDirection,
            });
        }
        return filtersWithCol;
    }
    FilterHelper.getFilters = getFilters;
    /**
     * Set the filters for an instance
     *
     */
    async function setFilters(instance, filters, externallyValidate = true) {
        var _a, _b, _c, _d, _e;
        delete instance.filters;
        const _modelName = (0, modelName_1.modelName)(instance);
        if (externallyValidate)
            await validateFilters(instance, filters);
        const existingFilters = await getFilters(instance);
        const areEqual = filtersAreEqual(filters, existingFilters);
        if (areEqual) {
            if (typeof (instance === null || instance === void 0 ? void 0 : instance["afterSetFilters"]) === "function") {
                await instance["afterSetFilters"](false);
            }
            return;
        }
        await Filter_1.Filter.destroy({
            where: { ownerId: instance.id, ownerType: _modelName },
        });
        const newPropertyFilters = [];
        for (const i in filters) {
            const filter = filters[i];
            const propertyFilter = await Filter_1.Filter.create({
                ownerId: instance.id,
                ownerType: _modelName,
                position: parseInt(i) + 1,
                key: filter.key,
                op: filter.op,
                match: (_b = (_a = filter.match) === null || _a === void 0 ? void 0 : _a.toString()) !== null && _b !== void 0 ? _b : null,
                relativeMatchNumber: (_c = filter.relativeMatchNumber) !== null && _c !== void 0 ? _c : null,
                relativeMatchUnit: (_d = filter.relativeMatchUnit) !== null && _d !== void 0 ? _d : null,
                relativeMatchDirection: (_e = filter.relativeMatchDirection) !== null && _e !== void 0 ? _e : null,
            });
            newPropertyFilters.push(propertyFilter);
        }
        instance.filters = newPropertyFilters;
        await instance.touch();
        if (typeof (instance === null || instance === void 0 ? void 0 : instance["afterSetFilters"]) === "function") {
            await instance["afterSetFilters"](true);
        }
    }
    FilterHelper.setFilters = setFilters;
    /**
     * Validate filters for an instance
     *
     */
    async function validateFilters(instance, filters) {
        if (!filters)
            filters = await getFilters(instance);
        const _pluginFilterOptions = await pluginFilterOptions(instance);
        for (const i in filters) {
            const filter = filters[i];
            const relevantOption = _pluginFilterOptions.filter((pfo) => pfo.key === filter.key)[0];
            if (!relevantOption) {
                throw new Error(`${filter.key} is not filterable`);
            }
            if (!relevantOption.ops.includes(filter.op)) {
                throw new Error(`"${filter.op}" cannot be applied to ${filter.key}`);
            }
        }
    }
    FilterHelper.validateFilters = validateFilters;
    /**
     * Get the options for an instance's Filter from its plugin
     */
    async function pluginFilterOptions(instance) {
        const { pluginConnection } = await instance.getPlugin();
        if (!pluginConnection.methods.sourceFilters)
            return [];
        const instanceOptions = await instance.getOptions(true);
        const source = await Source_1.Source.findOne({
            where: { id: instance.sourceId },
            include: [Option_1.Option, Mapping_1.Mapping],
        });
        const sourceOptions = await source.getOptions(true);
        const sourceMapping = await source.getMapping();
        const app = await source.$get("app", { include: [Option_1.Option], scope: null });
        const connection = await app.getConnection();
        const appOptions = await app.getOptions(true);
        const method = pluginConnection.methods.sourceFilters;
        const options = await method({
            connection,
            app,
            appId: app.id,
            appOptions,
            source,
            sourceId: source.id,
            sourceOptions,
            sourceMapping,
            property: instance instanceof Property_1.Property ? instance : undefined,
            propertyId: instance instanceof Property_1.Property ? instance.id : undefined,
            propertyOptions: instance instanceof Property_1.Property ? instanceOptions : undefined,
            schedule: instance instanceof Schedule_1.Schedule ? instance : undefined,
            scheduleId: instance instanceof Schedule_1.Schedule ? instance.id : undefined,
            scheduleOptions: instance instanceof Schedule_1.Schedule ? instanceOptions : undefined,
        });
        return options;
    }
    FilterHelper.pluginFilterOptions = pluginFilterOptions;
    /**
     * Determine if these Filters are equal
     */
    function filtersAreEqual(oldFilters, newFilters) {
        if (oldFilters.length !== newFilters.length)
            return false;
        if (oldFilters.length === newFilters.length && newFilters.length === 0) {
            return true;
        }
        function nullish(value) {
            if (value === null)
                return null;
            if (value === undefined)
                return null;
            if (value === "null")
                return null;
            return value.toString();
        }
        for (const i in oldFilters) {
            const A = oldFilters[i];
            const B = newFilters[i];
            if (A.key !== B.key)
                return false;
            if (A.op !== B.op)
                return false;
            if (nullish(A.match) !== nullish(B.match))
                return false;
            if (nullish(A.relativeMatchNumber) !== nullish(B.relativeMatchNumber)) {
                return false;
            }
            if (nullish(A.relativeMatchUnit) !== nullish(B.relativeMatchUnit)) {
                return false;
            }
            if (nullish(A.relativeMatchDirection) !== nullish(B.relativeMatchDirection)) {
                return false;
            }
        }
        return true;
    }
    FilterHelper.filtersAreEqual = filtersAreEqual;
})(FilterHelper = exports.FilterHelper || (exports.FilterHelper = {}));
