"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.MappingHelper = void 0;
const sequelize_1 = require("sequelize");
const Mapping_1 = require("../models/Mapping");
const Property_1 = require("../models/Property");
const Source_1 = require("./../models/Source");
const lockableHelper_1 = require("./lockableHelper");
const modelName_1 = require("./modelName");
const propertiesCache_1 = require("./caches/propertiesCache");
var MappingHelper;
(function (MappingHelper) {
    async function getMapping(instance, propertyColumn) {
        var _a;
        const mappingObject = {};
        const mappings = (_a = instance.mappings) !== null && _a !== void 0 ? _a : (await Mapping_1.Mapping.findAll({
            where: {
                ownerId: instance.id,
                ownerType: (0, modelName_1.modelName)(instance),
            },
        }));
        if (!instance.mappings)
            instance.mappings = mappings;
        for (const i in mappings) {
            const mapping = mappings[i];
            const property = await propertiesCache_1.PropertiesCache.findOneWithCache(mapping.propertyId, instance.modelId, "ready");
            if (!property) {
                throw new Error(`cannot find property or this source/destination not ready (remoteKey: ${mapping.remoteKey}, propertyId: ${mapping.propertyId})`);
            }
            mappingObject[mapping.remoteKey] = property[propertyColumn || "key"];
        }
        return mappingObject;
    }
    MappingHelper.getMapping = getMapping;
    async function getConfigMapping(instance) {
        const mappingObject = {};
        const mappings = await Mapping_1.Mapping.findAll({
            where: { ownerId: instance.id },
            include: [Property_1.Property],
        });
        for (const mapping of mappings) {
            const property = await mapping.property;
            if (!property) {
                throw new Error(`cannot find property or this source/destination not ready (remoteKey: ${mapping.remoteKey})`);
            }
            mappingObject[mapping.remoteKey] = property.getConfigId();
        }
        return mappingObject;
    }
    MappingHelper.getConfigMapping = getConfigMapping;
    async function setMapping(instance, mappings, externallyValidate = true) {
        delete instance.mappings;
        await lockableHelper_1.LockableHelper.beforeUpdateOptions(instance);
        await Mapping_1.Mapping.destroy({
            where: {
                ownerId: instance.id,
                ownerType: (0, modelName_1.modelName)(instance),
            },
        });
        const otherSourcePrimaryKey = instance instanceof Source_1.Source
            ? await Property_1.Property.findOne({
                include: {
                    model: Source_1.Source,
                    required: true,
                    where: {
                        modelId: instance.modelId,
                    },
                },
                where: {
                    isPrimaryKey: true,
                    sourceId: {
                        [sequelize_1.Op.ne]: instance.id,
                    },
                },
            })
            : undefined;
        let newMappings = [];
        const keys = Object.keys(mappings);
        for (const i in keys) {
            const remoteKey = keys[i];
            const key = mappings[remoteKey];
            const property = await propertiesCache_1.PropertiesCache.findOneWithCache(key, undefined, "ready", "key");
            if (!property)
                throw new Error(`cannot find property ${key}`);
            if (instance instanceof Source_1.Source && property.isArray) {
                throw new Error(`Sources cannot map to an array Property ${property.key} (${property.id})`);
            }
            const source = await property.$get("source", { scope: null });
            if (source.modelId !== instance.modelId) {
                throw new Error(`cannot map a ${instance.modelId} model to a ${source.modelId} property (${source.name})`);
            }
            if (otherSourcePrimaryKey && property.sourceId === instance.id) {
                throw new Error(`'${instance.name}' cannot map '${remoteKey}' to own Property '${key}'.
          '${otherSourcePrimaryKey.source.name}' is the primary Source for Model
          ${instance.modelId} and '${instance.name}' requires a mapping back to '${otherSourcePrimaryKey.source.name}.'`);
            }
            if (externallyValidate && instance instanceof Source_1.Source)
                await validateRemoteKey(instance, remoteKey);
            const mapping = await Mapping_1.Mapping.create({
                ownerId: instance.id,
                ownerType: (0, modelName_1.modelName)(instance),
                propertyId: property.id,
                remoteKey,
            });
            newMappings.push(mapping);
        }
        instance.mappings = newMappings;
        await instance.touch();
        // if there's an afterSetMapping hook and we want to commit our changes
        if (typeof instance["afterSetMapping"] === "function") {
            await instance["afterSetMapping"]();
        }
    }
    MappingHelper.setMapping = setMapping;
    async function validateRemoteKey(instance, remoteKey) {
        const previewAvailable = await instance.previewAvailable();
        if (!previewAvailable)
            return;
        const sourcePreview = await instance.sourcePreview();
        if (sourcePreview.length > 0) {
            const validKeys = Object.keys(sourcePreview[0]);
            if (!validKeys.includes(remoteKey)) {
                throw new Error(`"${remoteKey}" is not a valid remote mapping key for ${(0, modelName_1.modelName)(instance)} ${instance.name}`);
            }
        }
    }
})(MappingHelper = exports.MappingHelper || (exports.MappingHelper = {}));
