"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.AppRefreshQueryOps = void 0;
const Source_1 = require("../../models/Source");
const Schedule_1 = require("../../models/Schedule");
const Run_1 = require("../../models/Run");
var AppRefreshQueryOps;
(function (AppRefreshQueryOps) {
    /**
     * Query the app
     */
    async function runAppQuery(appRefreshQuery, testQuery) {
        const app = await appRefreshQuery.$get("app");
        const options = await app.getOptions();
        const { pluginApp } = await app.getPlugin();
        const connection = await app.getConnection();
        if (typeof pluginApp.methods.appQuery !== "function") {
            throw new Error(`app ${app.name} (${app.id}) of type ${app.type} cannot use app refresh query`);
        }
        const responseRows = await pluginApp.methods.appQuery({
            app,
            appId: app.id,
            appOptions: options,
            connection,
            refreshQuery: testQuery || appRefreshQuery.refreshQuery,
        });
        const sampleValue = JSON.stringify(Array.isArray(responseRows) ? responseRows[0] : responseRows);
        return sampleValue;
    }
    AppRefreshQueryOps.runAppQuery = runAppQuery;
    /**
     * Trigger all schedules for all sources associated with this app.
     * Optionally, cancel existing runs on those schedules (used by actions)
     */
    async function triggerSchedules(appRefreshQuery, stopRuns) {
        const sources = await Source_1.Source.findAll({
            where: { appId: appRefreshQuery.appId },
        });
        const schedulesToRun = [];
        for (const source of sources) {
            schedulesToRun.push(...(await Schedule_1.Schedule.findAll({
                where: {
                    sourceId: source.id,
                    refreshEnabled: true,
                    state: "ready",
                },
            })));
        }
        let runs = [];
        for (const schedule of schedulesToRun) {
            if (stopRuns === true) {
                const runningRun = await Run_1.Run.findOne({
                    where: { creatorId: schedule.id, state: "running" },
                });
                if (runningRun)
                    await runningRun.stop();
            }
            const run = await schedule.enqueueRun();
            runs.push(run);
        }
        return runs;
    }
    AppRefreshQueryOps.triggerSchedules = triggerSchedules;
    /**
     * Determine if it is time to run
     */
    async function shouldRun(appRefreshQuery) {
        if (appRefreshQuery.state !== "ready")
            return false;
        if (appRefreshQuery.lastConfirmedAt == null)
            return true;
        const delta = new Date().getTime() - appRefreshQuery.lastConfirmedAt.getTime();
        return delta > appRefreshQuery.recurringFrequency;
    }
    AppRefreshQueryOps.shouldRun = shouldRun;
    /**
     * Test the appRefreshQuery and return the sample value
     */
    async function test(appRefreshQuery, testQuery) {
        let success = false;
        let message;
        let error;
        try {
            const sampleValue = await AppRefreshQueryOps.runAppQuery(appRefreshQuery, testQuery);
            message = sampleValue;
            success = true;
        }
        catch (err) {
            error = err.message;
            success = false;
        }
        return {
            success,
            message,
            error,
        };
    }
    AppRefreshQueryOps.test = test;
})(AppRefreshQueryOps = exports.AppRefreshQueryOps || (exports.AppRefreshQueryOps = {}));
