"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ExportOps = void 0;
const sequelize_1 = require("sequelize");
const recordProperty_1 = require("../../modules/ops/recordProperty");
const Export_1 = require("../../models/Export");
const cls_1 = require("../../modules/cls");
const Option_1 = require("../../models/Option");
var ExportOps;
(function (ExportOps) {
    const defaultExportProcessingDelay = 1000 * 60 * 5;
    /** Count up the exports in each state, optionally filtered for a record or destination */
    async function totals(where = {}) {
        const totals = {
            draft: 0,
            pending: 0,
            processing: 0,
            canceled: 0,
            failed: 0,
            complete: 0,
        };
        for (const state of Export_1.ExportStates) {
            totals[state] = await Export_1.Export.count({
                where: Object.assign({}, where, { state }),
            });
        }
        return totals;
    }
    ExportOps.totals = totals;
    /**
     * Given an export with stringified old/new record properties, this method will re-'inflate' them, ie turning date strings back to date objects.
     * To be used in the getter, this method cannot be async.
     */
    function deserializeExportRecordProperties(serializedStringifiedProperties) {
        var _a, _b;
        const response = {};
        const serializedProperties = serializedStringifiedProperties
            ? JSON.parse(serializedStringifiedProperties)
            : {};
        for (const key in serializedProperties) {
            const type = (_a = serializedProperties[key]) === null || _a === void 0 ? void 0 : _a.type;
            const rawValue = (_b = serializedProperties[key]) === null || _b === void 0 ? void 0 : _b.rawValue;
            if (!type) {
                // legacy formatting
                response[key] = serializedProperties[key];
            }
            else {
                // current formatting
                if (Array.isArray(rawValue)) {
                    response[key] = rawValue.map((rv) => recordProperty_1.RecordPropertyOps.getValue(rv, type));
                }
                else {
                    response[key] = recordProperty_1.RecordPropertyOps.getValue(rawValue, type);
                }
            }
        }
        return response;
    }
    ExportOps.deserializeExportRecordProperties = deserializeExportRecordProperties;
    async function processPendingExportsForDestination(destination, limit = 100, delayMs = defaultExportProcessingDelay) {
        if (!delayMs || delayMs < defaultExportProcessingDelay) {
            delayMs = defaultExportProcessingDelay;
        }
        const app = await destination.$get("app", {
            scope: null,
            include: [Option_1.Option],
        });
        const { pluginConnection } = await destination.getPlugin();
        let _exports;
        _exports = await Export_1.Export.findAll({
            where: {
                state: "pending",
                destinationId: destination.id,
                sendAt: { [sequelize_1.Op.lte]: new Date() },
                startedAt: {
                    [sequelize_1.Op.or]: [null, { [sequelize_1.Op.lt]: new Date().getTime() - delayMs }],
                },
            },
            limit,
        });
        if (_exports.length > 0) {
            await Export_1.Export.update({ startedAt: new Date() }, {
                where: { id: { [sequelize_1.Op.in]: _exports.map((e) => e.id) } },
            });
            if (pluginConnection.methods.exportRecords) {
                // the plugin has a batch exportRecords method
                await cls_1.CLS.enqueueTask("export:sendBatch", {
                    destinationId: destination.id,
                    exportIds: _exports.map((e) => e.id),
                }, `exports:${app.type}`);
            }
            else {
                // the plugin has a per-record exportRecord method
                for (const _export of _exports) {
                    await cls_1.CLS.enqueueTask("export:send", {
                        destinationId: destination.id,
                        exportId: _export.id,
                    }, `exports:${app.type}`);
                }
            }
        }
        return _exports.length;
    }
    ExportOps.processPendingExportsForDestination = processPendingExportsForDestination;
    /**
     * Mark an array of Exports complete
     */
    async function completeBatch(_exports) {
        if (_exports.length === 0)
            return;
        const now = new Date();
        await Export_1.Export.update({
            errorMessage: null,
            errorLevel: null,
            completedAt: now,
            state: "complete",
        }, {
            where: { id: { [sequelize_1.Op.in]: _exports.map((e) => e.id) } },
            returning: true,
        });
        for (const _export of _exports) {
            _export.state = "complete";
            _export.completedAt = now;
            await Export_1.Export.logExport(_export);
        }
    }
    ExportOps.completeBatch = completeBatch;
    /**
     * Sets export(s) to pending
     */
    async function retryExport(where, saveExports) {
        if (saveExports) {
            const [count] = await Export_1.Export.update({
                state: "pending",
                sendAt: new Date(),
                startedAt: null,
                completedAt: null,
                errorMessage: null,
                errorLevel: null,
                retryCount: 0,
            }, { where });
            return count;
        }
        return Export_1.Export.count({ where });
    }
    /**
     * Finds failed exports within a time range and sets them back to pending
     */
    async function retryFailedExports(startDate, endDate, destination, saveExports = true) {
        const where = {
            state: "failed",
            createdAt: {
                [sequelize_1.Op.gte]: startDate,
                [sequelize_1.Op.lte]: endDate,
            },
        };
        if (destination)
            where.destinationId = destination.id;
        return retryExport(where, saveExports);
    }
    ExportOps.retryFailedExports = retryFailedExports;
    /**
     * Finds finds and retries a failed or canceled export by id
     */
    async function retryExportById(exportId) {
        const where = {
            state: { [sequelize_1.Op.in]: ["failed", "canceled"] },
            id: exportId,
        };
        const count = await retryExport(where, true);
        if (count === 0) {
            throw new Error("No export found");
        }
        return count;
    }
    ExportOps.retryExportById = retryExportById;
})(ExportOps = exports.ExportOps || (exports.ExportOps = {}));
